@@ -28,7 +28,9 @@
 #include "kpassivepopup.h"
 #include "knotifyconfig.h"
 #include "knotification.h"
+#include "debug_p.h"
 
+#include <QApplication>
 #include <QBuffer>
 #include <QImage>
 #include <QLabel>
@@ -42,7 +44,6 @@
 #include <QDBusError>
 #include <QDBusMessage>
 #include <QXmlStreamReader>
-#include <QDebug>
 #include <QMap>
 #include <QHash>
 #include <QXmlStreamEntityResolver>
@@ -52,6 +53,7 @@
 #include <QScreen>
 #include <QFontMetrics>
 #include <QIcon>
+#include <QUrl>
 
 #include <kconfiggroup.h>
 #include <KCodecs/KCharsets>
@@ -107,6 +109,9 @@ public:
      * Specifies if DBus Notifications interface exists on session bus
      */
     bool dbusServiceExists;
+
+    bool dbusServiceActivatable;
+
     /**
      * DBus notification daemon capabilities cache.
      * Do not use this variable. Use #popupServerCapabilities() instead.
@@ -159,6 +164,7 @@ NotifyByPopup::NotifyByPopup(QObject *parent)
 {
     d->animationTimer = 0;
     d->dbusServiceExists = false;
+    d->dbusServiceActivatable = false;
     d->dbusServiceCapCacheDirty = true;
     d->nextPosition = -1;
 
@@ -167,7 +173,7 @@ NotifyByPopup::NotifyByPopup(QObject *parent)
     d->dbusServiceExists = interface && interface->isServiceRegistered(dbusServiceName);
 
     if (d->dbusServiceExists) {
-        onServiceOwnerChanged(dbusServiceName, QString(), "_"); //connect signals
+        onServiceOwnerChanged(dbusServiceName, QString(), QStringLiteral("_")); //connect signals
     }
 
     // to catch register/unregister events from service in runtime
@@ -178,34 +184,20 @@ NotifyByPopup::NotifyByPopup(QObject *parent)
     connect(watcher, SIGNAL(serviceOwnerChanged(QString,QString,QString)),
             SLOT(onServiceOwnerChanged(QString,QString,QString)));
 
+#ifndef Q_WS_WIN
     if (!d->dbusServiceExists) {
-        bool startfdo = false;
-#ifdef Q_WS_WIN
-        startfdo = true;
-#else
-        if (qgetenv("KDE_FULL_SESSION").isEmpty()) {
-            QDBusMessage message = QDBusMessage::createMethodCall("org.freedesktop.DBus",
-                                                                  "/org/freedesktop/DBus",
-                                                                  "org.freedesktop.DBus",
-                                                                  "ListActivatableNames");
-
-            // FIXME - this should be async
-            QDBusReply<QStringList> reply = QDBusConnection::sessionBus().call(message);
-            if (reply.isValid() && reply.value().contains(dbusServiceName)) {
-                startfdo = true;
-                // We need to set d->dbusServiceExists to true because dbus might be too slow
-                // starting the service and the first call to NotifyByPopup::notify
-                // might not have had the service up, by setting this to true we
-                // guarantee it will still go through dbus and dbus will do the correct
-                // thing and wait for the service to go up
-                d->dbusServiceExists = true;
-            }
-        }
-#endif
-        if (startfdo) {
-            QDBusConnection::sessionBus().interface()->startService(dbusServiceName);
+        QDBusMessage message = QDBusMessage::createMethodCall(QStringLiteral("org.freedesktop.DBus"),
+                                                                QStringLiteral("/org/freedesktop/DBus"),
+                                                                QStringLiteral("org.freedesktop.DBus"),
+                                                                QStringLiteral("ListActivatableNames"));
+        QDBusReply<QStringList> reply = QDBusConnection::sessionBus().call(message);
+        if (reply.isValid() && reply.value().contains(dbusServiceName)) {
+            d->dbusServiceActivatable = true;
+            //if the service is activatable, we can assume it exists even if it is not currently running
+            d->dbusServiceExists = true;
         }
     }
+#endif
 }
 
 
@@ -270,7 +262,7 @@ void NotifyByPopup::notify(KNotification *notification, const KNotifyConfig &not
 
         // The first arg is QPixmap*, however that pixmap is not used
         // at all (it has Q_UNUSED) so just set it to 0
-        NotifyByPopupGrowl::popup(Q_NULLPTR, timeout, appCaption, text);
+        NotifyByPopupGrowl::popup(nullptr, timeout, appCaption, text);
 
         // Finish immediately, because current NotifyByPopupGrowl can't callback
         finish(notification);
@@ -280,7 +272,12 @@ void NotifyByPopup::notify(KNotification *notification, const KNotifyConfig &not
     // Check if this object lives in the GUI thread and return if it doesn't
     // as Qt cannot create/handle widgets in non-GUI threads
     if (QThread::currentThread() != qApp->thread()) {
-        qWarning() << "KNotification did not detect any running org.freedesktop.Notifications server and fallback notifications cannot be used from non-GUI thread!";
+        qCWarning(LOG_KNOTIFICATIONS) << "KNotification did not detect any running org.freedesktop.Notifications server and fallback notifications cannot be used from non-GUI thread!";
+        return;
+    }
+
+    if (!qobject_cast<QApplication *>(QCoreApplication::instance())) {
+        qCWarning(LOG_KNOTIFICATIONS) << "KNotification did not detect any running org.freedesktop.Notifications server and fallback notifications cannot be used without a QApplication!";
         return;
     }
 
@@ -291,7 +288,7 @@ void NotifyByPopup::notify(KNotification *notification, const KNotifyConfig &not
 
     QRect screen = QGuiApplication::primaryScreen()->availableGeometry();
     if (d->nextPosition == -1) {
-        d->nextPosition = screen.y() + screen.height() - 32;
+        d->nextPosition = screen.top();
     }
     pop->setAutoDelete(true);
     connect(pop, SIGNAL(destroyed()), this, SLOT(onPassivePopupDestroyed()));
@@ -431,10 +428,12 @@ void NotifyByPopup::onServiceOwnerChanged(const QString &serviceName, const QStr
 
     d->dbusServiceCapCacheDirty = true;
     d->popupServerCapabilities.clear();
-    d->notificationQueue.clear();
 
     if (newOwner.isEmpty()) {
-        d->dbusServiceExists = false;
+        d->notificationQueue.clear();
+        if (!d->dbusServiceActivatable) {
+            d->dbusServiceExists = false;
+        }
     } else if (oldOwner.isEmpty()) {
         d->dbusServiceExists = true;
 
@@ -442,21 +441,21 @@ void NotifyByPopup::onServiceOwnerChanged(const QString &serviceName, const QStr
         bool connected = QDBusConnection::sessionBus().connect(QString(), // from any service
                                                                dbusPath,
                                                                dbusInterfaceName,
-                                                               "ActionInvoked",
+                                                               QStringLiteral("ActionInvoked"),
                                                                this,
                                                                SLOT(onGalagoNotificationActionInvoked(uint,QString)));
         if (!connected) {
-            qWarning() << "warning: failed to connect to ActionInvoked dbus signal";
+            qCWarning(LOG_KNOTIFICATIONS) << "warning: failed to connect to ActionInvoked dbus signal";
         }
 
         connected = QDBusConnection::sessionBus().connect(QString(), // from any service
                                                           dbusPath,
                                                           dbusInterfaceName,
-                                                          "NotificationClosed",
+                                                          QStringLiteral("NotificationClosed"),
                                                           this,
                                                           SLOT(onGalagoNotificationClosed(uint,uint)));
         if (!connected) {
-            qWarning() << "warning: failed to connect to NotificationClosed dbus signal";
+            qCWarning(LOG_KNOTIFICATIONS) << "warning: failed to connect to NotificationClosed dbus signal";
         }
     }
 }
@@ -470,10 +469,11 @@ void NotifyByPopup::onGalagoNotificationActionInvoked(uint notificationId, const
 
     KNotification *n = *iter;
     if (n) {
-        emit actionInvoked(n->id(), actionKey.toUInt());
-        // now close notification - similar to popup behaviour
-        // (popups are hidden after link activation - see 'connects' of linkActivated signal above)
-        d->closeGalagoNotification(n);
+        if (actionKey == QStringLiteral("default")) {
+            emit actionInvoked(n->id(), 0);
+        } else {
+            emit actionInvoked(n->id(), actionKey.toUInt());
+        }
     } else {
         d->galagoNotifications.erase(iter);
     }
@@ -507,7 +507,7 @@ void NotifyByPopup::onGalagoServerReply(QDBusPendingCallWatcher *watcher)
     watcher->deleteLater();
     KNotification *notification = watcher->property("notificationObject").value<KNotification*>();
     if (!notification) {
-        qWarning() << "Invalid notification object passed in DBus reply watcher; notification will probably break";
+        qCWarning(LOG_KNOTIFICATIONS) << "Invalid notification object passed in DBus reply watcher; notification will probably break";
         return;
     }
 
@@ -531,10 +531,10 @@ void NotifyByPopup::onGalagoServerCapabilitiesReceived(const QStringList &capabi
 
 void NotifyByPopupPrivate::getAppCaptionAndIconName(const KNotifyConfig &notifyConfig, QString *appCaption, QString *iconName)
 {
-    KConfigGroup globalgroup(&(*notifyConfig.eventsfile), QString("Global"));
+    KConfigGroup globalgroup(&(*notifyConfig.eventsfile), QStringLiteral("Global"));
     *appCaption = globalgroup.readEntry("Name", globalgroup.readEntry("Comment", notifyConfig.appname));
 
-    KConfigGroup eventGroup(&(*notifyConfig.eventsfile), QString("Event/%1").arg(notifyConfig.eventid));
+    KConfigGroup eventGroup(&(*notifyConfig.eventsfile), QStringLiteral("Event/%1").arg(notifyConfig.eventid));
     if (eventGroup.hasKey("IconName")) {
         *iconName = eventGroup.readEntry("IconName", notifyConfig.appname);
     } else {
@@ -593,14 +593,14 @@ void NotifyByPopupPrivate::fillPopup(KPassivePopup *popup, KNotification *notifi
 
 
     if (!notification->actions().isEmpty()) {
-        QString linkCode = QString::fromLatin1("<p align=\"right\">");
+        QString linkCode = QStringLiteral("<p align=\"right\">");
         int i = 0;
         Q_FOREACH (const QString &it, notification->actions()) {
             i++;
-            linkCode += QString::fromLatin1("&nbsp;<a href=\"%1/%2\">%3</a>").arg(QString::number(notification->id()), QString::number(i), it.toHtmlEscaped());
+            linkCode += QStringLiteral("&nbsp;<a href=\"%1/%2\">%3</a>").arg(QString::number(notification->id()), QString::number(i), it.toHtmlEscaped());
         }
 
-        linkCode += QString::fromLatin1("</p>");
+        linkCode += QLatin1String("</p>");
         QLabel *link = new QLabel(linkCode , vb );
         link->setTextInteractionFlags(Qt::LinksAccessibleByMouse);
         link->setOpenExternalLinks(false);
@@ -624,7 +624,7 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
         }
     }
 
-    QDBusMessage dbusNotificationMessage = QDBusMessage::createMethodCall(dbusServiceName, dbusPath, dbusInterfaceName, "Notify");
+    QDBusMessage dbusNotificationMessage = QDBusMessage::createMethodCall(dbusServiceName, dbusPath, dbusInterfaceName, QStringLiteral("Notify"));
 
     QList<QVariant> args;
 
@@ -644,11 +644,11 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
     QString title = notification->title().isEmpty() ? appCaption : notification->title();
     QString text = notification->text();
 
-    if (!popupServerCapabilities.contains("body-markup")) {
-        if (title.startsWith("<html>")) {
+    if (!popupServerCapabilities.contains(QStringLiteral("body-markup"))) {
+        if (title.startsWith(QLatin1String("<html>"))) {
             title = stripHtml(title);
         }
-        if (text.startsWith("<html>")) {
+        if (text.startsWith(QLatin1String("<html>"))) {
             text = stripHtml(text);
         }
     }
@@ -662,7 +662,12 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
     // assign id's to actions like it's done in fillPopup() method
     // (i.e. starting from 1)
     QStringList actionList;
-    if (popupServerCapabilities.contains("actions")) {
+    if (popupServerCapabilities.contains(QStringLiteral("actions"))) {
+        QString defaultAction = notification->defaultAction();
+        if (!defaultAction.isEmpty()) {
+            actionList.append(QStringLiteral("default"));
+            actionList.append(defaultAction);
+        }
         int actId = 0;
         Q_FOREACH (const QString &actionName, notification->actions()) {
             actId++;
@@ -678,9 +683,31 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
     // According to fdo spec, the app_name is supposed to be the applicaton's "pretty name"
     // but in some places it's handy to know the application name itself
     if (!notification->appName().isEmpty()) {
-        hintsMap["x-kde-appname"] = notification->appName();
+        hintsMap[QStringLiteral("x-kde-appname")] = notification->appName();
     }
 
+    if (!notification->eventId().isEmpty()) {
+        hintsMap[QStringLiteral("x-kde-eventId")] = notification->eventId();
+    }
+
+    if (notification->flags() & KNotification::SkipGrouping) {
+        hintsMap[QStringLiteral("x-kde-skipGrouping")] = 1;
+    }
+
+    if (!notification->urls().isEmpty()) {
+        hintsMap[QStringLiteral("x-kde-urls")] = QUrl::toStringList(notification->urls());
+    }
+
+    if (!(notification->flags() & KNotification::Persistent)) {
+        hintsMap[QStringLiteral("transient")] = true;
+    }
+
+#if QT_VERSION >= QT_VERSION_CHECK(5, 7, 0)
+    if (!qApp->desktopFileName().isEmpty()) {
+        hintsMap[QStringLiteral("desktop-entry")] = qApp->desktopFileName();
+    }
+#endif
+
     //FIXME - reenable/fix
     // let's see if we've got an image, and store the image in the hints map
     if (!notification->pixmap().isNull()) {
@@ -689,7 +716,7 @@ bool NotifyByPopupPrivate::sendNotificationToGalagoServer(KNotification *notific
         buffer.open(QIODevice::WriteOnly);
         notification->pixmap().save(&buffer, "PNG");
         buffer.close();
-        hintsMap["image_data"] = ImageConverter::variantForImage(QImage::fromData(pixmapData));
+        hintsMap[QStringLiteral("image_data")] = ImageConverter::variantForImage(QImage::fromData(pixmapData));
     }
 
     args.append(hintsMap); // hints
@@ -719,12 +746,12 @@ void NotifyByPopupPrivate::closeGalagoNotification(KNotification *notification)
     uint galagoId = galagoNotifications.key(notification, 0);
 
     if (galagoId == 0) {
-        qDebug() << "not found dbus id to close" << notification->id();
+        qCDebug(LOG_KNOTIFICATIONS) << "not found dbus id to close" << notification->id();
         return;
     }
 
     QDBusMessage m = QDBusMessage::createMethodCall(dbusServiceName, dbusPath,
-                                                    dbusInterfaceName, "CloseNotification");
+                                                    dbusInterfaceName, QStringLiteral("CloseNotification"));
     QList<QVariant> args;
     args.append(galagoId);
     m.setArguments(args);
@@ -733,9 +760,8 @@ void NotifyByPopupPrivate::closeGalagoNotification(KNotification *notification)
     bool queued = QDBusConnection::sessionBus().send(m);
 
     if (!queued) {
-        qWarning() << "Failed to queue dbus message for closing a notification";
+        qCWarning(LOG_KNOTIFICATIONS) << "Failed to queue dbus message for closing a notification";
     }
-
 }
 
 void NotifyByPopupPrivate::queryPopupServerCapabilities()
@@ -745,8 +771,8 @@ void NotifyByPopupPrivate::queryPopupServerCapabilities()
             popupServerCapabilities = NotifyByPopupGrowl::capabilities();
         } else {
             // Return capabilities of the KPassivePopup implementation
-            popupServerCapabilities = QStringList() << "actions" << "body" << "body-hyperlinks"
-                                                      << "body-markup" << "icon-static";
+            popupServerCapabilities = QStringList() << QStringLiteral("actions") << QStringLiteral("body") << QStringLiteral("body-hyperlinks")
+                                                      << QStringLiteral("body-markup") << QStringLiteral("icon-static");
         }
     }
 
@@ -754,12 +780,12 @@ void NotifyByPopupPrivate::queryPopupServerCapabilities()
         QDBusMessage m = QDBusMessage::createMethodCall(dbusServiceName,
                                                         dbusPath,
                                                         dbusInterfaceName,
-                                                        "GetCapabilities");
+                                                        QStringLiteral("GetCapabilities"));
 
         QDBusConnection::sessionBus().callWithCallback(m,
                                                        q,
                                                        SLOT(onGalagoServerCapabilitiesReceived(QStringList)),
-                                                       0,
+                                                       nullptr,
                                                        -1);
     }
 }
@@ -781,7 +807,7 @@ QString NotifyByPopupPrivate::stripHtml(const QString &text)
 
     if (r.hasError()) {
         // XML error in the given text, just return the original string
-        qWarning() << "Notification to send to backend which does "
+        qCWarning(LOG_KNOTIFICATIONS) << "Notification to send to backend which does "
                          "not support HTML, contains invalid XML:"
                       << r.errorString() << "line" << r.lineNumber()
                       << "col" << r.columnNumber();
@@ -802,7 +828,7 @@ QString NotifyByPopupPrivate::HtmlEntityResolver::resolveUndeclaredEntity(const
     QChar ent = KCharsets::fromEntity('&' + name);
 
     if (ent.isNull()) {
-        qWarning() << "Notification to send to backend which does "
+        qCWarning(LOG_KNOTIFICATIONS) << "Notification to send to backend which does "
                       "not support HTML, contains invalid entity: "
                    << name;
         ent = ' ';
